/***************************************************************************\
  "icm.h" - Intermediate Code Manager (Phase D)
  Revision 12; 10 Nov 2002 - 30 Jan 2003
  Copyright (C) 2002-2003  Kevin P. Albrecht
\***************************************************************************/

#ifndef ICM_H
#define ICM_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdlib.h>

// Local //
#include "global.h"
#include "error.h"
#include "table.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// External //

typedef enum // Three-Address Code Operations
{
  tacNOP,
  tacLINE, // marks a statement holding a linenumber
  tacJUMP, tacSTOP, tacCOND, tacCALL, tacRET,
  tacXIR, tacXIS, tacXRI, tacXRS, tacXSI, tacXSR,
  tacSET, tacPUSHL, tacPOPL, tacPUSHR, tacPOPR,
  tacADD, tacSUB, tacMULT, tacDIV, tacMOD, tacPOW,
  tacNEG, tacRAND,
  tacEQ, tacNEQ, tacLT, tacGT, tacLTEQ, tacGTEQ,
  tacOR, tacAND,
  tacCAT, tacSEQ, tacSNEQ,
  tacCIN, tacCOUT,
  tacASC, tacFIO,
} TAC;

char *gvsTACStrings[] =
{
  "nop",
  "line",
  "jump", "stop", "cond", "call", "ret",
  "xir", "xis", "xri", "xrs", "xsi", "xsr",
  "set", "pushl", "popl", "pushr", "popr",
  "add", "sub", "mult", "div", "mod", "pow",
  "neg", "rand",
  "eq", "neq", "lt", "gt", "lteq", "gteq",
  "or", "and",
  "cat", "seq", "sneq",
  "cin", "cout",
  "asc", "fio",
};

typedef struct
{
  TAC tac;
  Type type;
  Symbol *a;
  Symbol *b;
  Symbol *c;
  int number; // misc. number holder    tac         number
              //                        ===         ======
              //                        tacLINE     line number
} Statement;

Statement *statement_new (TAC aTac, Symbol *aA, Symbol *aB, Symbol *aC);
void statement_del (Statement *statement);

Statement** gvProgram; // vector of statements
int gxLast;            // index of last statement in the gvProgram vector
int gxCurr;            // index of current statement in the gvProgram vector

void icm_init ();
void icm_deinit ();
int icm_put (Statement *statement);
void icm_putline ();
void icm_reset ();
void icm_goto (int x);
Statement *icm_getnext ();
Statement *icm_get (int x);

// Internal //
bool mReset;

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// statement_new
Statement *statement_new (TAC aTac, Symbol *aA, Symbol *aB, Symbol *aC)
{
  Statement *temp;

  temp = (Statement *) JALLOC(sizeof (Statement));
  temp->tac = aTac;
  temp->a = aA;
  temp->b = aB;
  temp->c = aC;
  temp->number = 0;

  return temp;
}

//// statement_del
void statement_del (Statement *statement)
{
  if (statement)
  {
    // NOTE: Symbols are not deleted here; they are deleted in   //
    //       scope_deinit() by way of scope_add_symbol_to_list() //

    JFREE(statement);
  }
}

//// icm_init
void icm_init ()
{
  int i;

  gvProgram = (Statement **) JALLOC(sizeof (Statement *) * GLOBAL_ICM_MAX_INSTRUCTIONS);
  for (i = 0; i < GLOBAL_ICM_MAX_INSTRUCTIONS; i++)
  {
    gvProgram[i] = NULL;
  }

  gxLast = 0;
  gxCurr = 0;
  mReset = false;
}

//// icm_deinit
void icm_deinit ()
{
  int i;

  for (i = 0; i < GLOBAL_ICM_MAX_INSTRUCTIONS; i++)
  {
    if (gvProgram[i]) statement_del (gvProgram[i]);
  }
  JFREE(gvProgram);
}

//// icm_put
int icm_put (Statement *statement)
{
  if (mReset == false)
  {
    if ((gxCurr + 1) > GLOBAL_ICM_MAX_INSTRUCTIONS)
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    gvProgram[gxCurr] = statement;
    gxCurr++;
    gxLast++;
  }
  else
  {
    err (true, EI_WRITE_AFTER_RESET, NULL);
    erre (EFF_TERMINATE);
  }

  return (gxCurr - 1);
}

//// icm_putline
void icm_putline ()
{
  Statement *statement;
  Symbol *symbolA, *symbolB;
  char *sTemp;

  if (mReset == false)
  {
    if ((gxCurr + 1) > GLOBAL_ICM_MAX_INSTRUCTIONS)
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    if (!gbFast)
    {
      symbolA = sym_new (gsCurrentFilename);
      sym_set (symbolA, repLiteral, tStr, false, gsCurrentFilename);
      scope_add_symbol_to_list (symbolA); // add symbol to deletion list

      symbolB = sym_new (gsCurrentFilename);
      sTemp = k_trim(mBufferStackTop->text);
      sym_set (symbolB, repLiteral, tStr, false, sTemp);
      k_del (sTemp);
      scope_add_symbol_to_list (symbolB); // add symbol to deletion list
      
      statement = statement_new (tacLINE, symbolA, symbolB, NULL);
      statement->number = gnCurrentLine;
      gvProgram[gxCurr] = statement;
      gxCurr++;
      gxLast++;
    }
  }
  else
  {
    err (true, EI_WRITE_AFTER_RESET, NULL);
    erre (EFF_TERMINATE);
  }
}

//// icm_reset
void icm_reset ()
{
  gxCurr = 0;
  mReset = true;
}

//// icm_goto
void icm_goto (int x)
{
  if (mReset == true)
  {
    if ((x > gxLast) || (x > GLOBAL_ICM_MAX_INSTRUCTIONS))
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    gxCurr = x;
  }
  else
  {
    err (true, EI_GOTO_BEFORE_RESET, NULL);
    erre (EFF_TERMINATE);
  }
}

//// icm_getnext
Statement *icm_getnext ()
{
  if (mReset == true)
  {
    if ((gxCurr + 1) > gxLast)
    {
      return NULL; // there are no more statements, so return NULL
    }

    return gvProgram[gxCurr++];
  }
  else
  {
    err (true, EI_READ_BEFORE_RESET, NULL);
    erre (EFF_TERMINATE);
    return NULL;
  }
}

//// icm_get
Statement *icm_get (int x)
{
  if (mReset == true)
  {
    if ((x > gxLast) || (x > GLOBAL_ICM_MAX_INSTRUCTIONS))
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      erre (EFF_TERMINATE);
    }

    return gvProgram[x];
  }
  else
  {
    err (true, EI_READ_BEFORE_RESET, NULL);
    erre (EFF_TERMINATE);
    return NULL;
  }
}

#endif
